import  os
import re

from corpus import Token, Document, Corpus

import spacy
from spacy.tokens import Doc

nlp = spacy.load("en_core_web_sm")
# make spacy always break on special characters, and on digit sequences

class LiberalTokenizer:
    def __init__(self, vocab):
        self.vocab = vocab
    def __call__(self, text):
        words = []
        spaces = []
        segments = re.compile(r'\s').split(text)
        breakpoint
        #segments = text.split()
        for segment in segments:
            if segment == '':
                words.append(' ')
                spaces.append(False)
            else:
                for part in re.finditer(r'[A-Za-z]+|[0-9]+|[^A-Za-z0-9]', segment):
                    words.append(part.group())
                    spaces.append(False)
                spaces[-1] = True
        return Doc(self.vocab, words=words, spaces=spaces)

nlp.tokenizer = LiberalTokenizer(nlp.vocab)

dig = re.compile('[0-9]')
def normalize_text(s):
    #such a clean and beautiful interface...
    if s is None:
        return s
    s = s.lower()
    s = re.sub(dig, '0', s)
    s = s.replace("``", '"')
    s = s.replace("''", '"')
    return s


def load_corpus(path, restricted_features=False):
    documents = []
    document_tokens = []
    document_labels = []
    abstracts_path = path + '.abstracts.txt'
    docs = {}
    tokens = {}
    labels = {}
    impossible_spans = {}
    with open(abstracts_path) as f:
        for line in f:
            line = line.strip()
            doc_id, title, abstract = line.split('\t')
            assert doc_id not in docs
            title_doc = nlp(title)
            abstract_doc = nlp(abstract)
            docs[doc_id] = title_doc, abstract_doc
            labels[doc_id] = (['O'] * len(title_doc), ['O'] * len(abstract_doc))
            impossible_spans[doc_id] = []
    annotations_path = path + '.annotations.txt'
    with open(annotations_path) as f:
        for line in f:
            line = line.strip().split('\t')
            doc_id = line[0]
            tuple_idx = 'TA'.index(line[1])
            doc = docs[doc_id][tuple_idx]
            label_sequence = labels[doc_id][tuple_idx]
            start_idx = int(line[2])
            end_idx = int(line[3])
            if start_idx == end_idx:
                # annotation error -- empty span
                continue
            span_type = line[5]

            # in case the span doesn't respect token boundaries,
            # we save a backup so we can undo changes to this
            label_sequence_backup = label_sequence[:]
            
            inside = False
            for i, token in enumerate(doc):
                if inside:
                    if token.idx >= end_idx:
                        # bad boundary
                        labels[doc_id][tuple_idx][:] = label_sequence_backup
                        impossible_spans[doc_id].append(span_type)
                        break
                    label_sequence[i] = 'I-' + span_type
                else:
                    if token.idx + len(token) > start_idx:
                        inside = True
                        label_sequence[i] = 'B-' + span_type
                    elif token.idx > start_idx:
                        # bad boundary
                        labels[doc_id][tuple_idx][:] = label_sequence_backup
                        impossible_spans[doc_id].append(span_type)
                        break

                if inside and token.idx + len(token) >= end_idx:
                    if token.idx + len(token) > end_idx:
                        # bad boundary
                        labels[doc_id][tuple_idx][:] = label_sequence_backup
                        impossible_spans[doc_id].append(span_type)
                        break
                    break
    documents = []
    for doc_id in docs:
        document_labels = labels[doc_id][0] + ['O'] + labels[doc_id][1]
        document_tokens = []
        for tok in list(docs[doc_id][0]) + [None] + list(docs[doc_id][1]):
            if tok is None:
                token = Token('<separator>', [('separator',)])
            else:
                if restricted_features:
                    features = [
                        ('word', normalize_text(tok.text))
                    ]
                else:
                    features = [
                        ('word', normalize_text(tok.text)),
                        ('pos', tok.pos_),
                        ('lemma', tok.lemma_),
                        ('allCaps', tok.is_upper),
                        ('upperInitial', tok.text[0].isupper()),
                        ('alpha', tok.is_alpha),
                        ('digit', tok.is_digit)
                    ]
                token = Token(tok.text, features)
            document_tokens.append(token)
        document = Document(document_tokens, document_labels)
        document.missed_spans.extend(impossible_spans[doc_id])
        documents.append(document)

    return Corpus(documents)

if __name__ == '__main__':
    import sys
    corpus = load_corpus(sys.argv[1])
    breakpoint()
