import os
from collections import defaultdict
from task_properties import get_task_properties, include_task
import pickle

def get_final_results(path):
    with open(path) as f:
        lines = f.readlines()

    while 'FINAL RESULTS' not in lines[0]:
        lines = lines[1:]

    return ''.join(lines)

def parse_final_results(table_string):
    lines = table_string.split('\n')
    # the first 5 lines are garbage
    lines = lines[5:]
    rows = []
    for line in lines:
        if '+' in line:
            continue
        fields = line.split('|')
        row = []
        for field in fields:
            field = field.strip()
            if len(field) > 0:
                row.append(field)
        
        if len(row) == 0:
            continue
        row[1] = int(row[1])
        row[3] = int(row[2])
        row[2] = int(row[3])
        row[4] = float(row[4])
        row[5] = float(row[5])
        row[6] = float(row[6])
        rows.append(row)
    return rows

def get_average_f1s(n_trials=5):
    try:
        with open("cached_average_f1s.pkl", 'rb') as f:
            cached_f1s = pickle.load(f)
            return cached_f1s
    except FileNotFoundError:
        pass

    log_paths = [os.path.join('logs', fn) for fn in os.listdir('logs')]
    model_corpus_trials = defaultdict(int)
    f1s = []
    for path in log_paths:
        filename = os.path.basename(path)
        if 'test' in filename:
            continue
        try:
            final_results = get_final_results(path)
        except IndexError:
            continue

        model_properties = []
        # determine model properties:
        if 'bert' in filename:
            model_properties.append('BERT')
        if 'lstm' in filename:
            model_properties.append('LSTM')
        if 'crf' in filename:
            model_properties.append('CRF')
        if 'bfm' in filename:
            assert model_properties == []
            model_properties = ['BERT', 'LSTM', 'CRF']
        if 'LSTM' in model_properties or 'BERT' not in model_properties:
            if '-r-' not in filename:
                model_properties.append('Feat')
        if 'baseline' not in filename and (set(model_properties) & {'BERT', 'LSTM', 'CRF'} == set()):
            breakpoint()
        model_properties = tuple(model_properties)
        
        # determine the corpus
        if 'parc' in filename:
            corpus = 'parc'
        elif 'ontonotes' in filename:
            corpus = 'ontonotes'
        elif 'chemdner' in filename:
            corpus = 'chemdner'
        elif 'riqua' in filename:
            corpus = 'riqua'
        elif 'conll2k' in filename:
            corpus = 'conll2k'
        else:
            breakpoint()

        if model_corpus_trials[model_properties, corpus] >= n_trials:
            continue
        model_corpus_trials[model_properties, corpus] += 1

        table = parse_final_results(final_results)

        f1s.append((model_properties, corpus, table[-1][-1]))
        
    with open("cached_average_f1s.pkl", 'wb') as f:
        pickle.dump(f1s, f)

    return f1s

def get_f1s(n_trials=5):
    try:
        with open("cached_f1s.pkl", 'rb') as f:
            cached_f1s = pickle.load(f)
            return cached_f1s
    except FileNotFoundError:
        pass

    log_paths = [os.path.join('logs', fn) for fn in os.listdir('logs')]
    model_corpus_trials = defaultdict(int)
    f1s = []
    for path in log_paths:
        filename = os.path.basename(path)
        if 'test' in filename:
            continue
        try:
            final_results = get_final_results(path)
        except IndexError:
            continue

        model_properties = []
        # determine model properties:
        if 'bert' in filename:
            model_properties.append('BERT')
        if 'lstm' in filename:
            model_properties.append('LSTM')
        if 'crf' in filename:
            model_properties.append('CRF')
        if 'bfm' in filename:
            assert model_properties == []
            model_properties = ['BERT', 'LSTM', 'CRF']
        if 'LSTM' in model_properties or 'BERT' not in model_properties:
            if '-r-' not in filename:
                model_properties.append('Feat')
        if 'baseline' not in filename and (set(model_properties) & {'BERT', 'LSTM', 'CRF'} == set()):
            breakpoint()
        model_properties = tuple(model_properties)
        
        # determine the corpus
        if 'parc' in filename:
            corpus = 'parc'
        elif 'ontonotes' in filename:
            corpus = 'ontonotes'
        elif 'chemdner' in filename:
            corpus = 'chemdner'
        elif 'riqua' in filename:
            corpus = 'riqua'
        elif 'conll2k' in filename:
            corpus = 'conll2k'
        else:
            breakpoint()

        if model_corpus_trials[model_properties, corpus] >= n_trials:
            continue
        model_corpus_trials[model_properties, corpus] += 1

        table = parse_final_results(final_results)

        for row in table[:-1]: # exclude the final micro-averaged task
            task = row[0]
            if include_task(corpus, task):
                f1s.append((model_properties, corpus, task, row[-1]))

    with open("cached_f1s.pkl", 'wb') as f:
        pickle.dump(f1s, f)

    return f1s
