import os
import re

from corpus import Token, Document, Corpus

dig = re.compile('[0-9]')
def normalize_text(s):
    #such a clean and beautiful interface...
    if s is None:
        return s
    s = s.lower()
    s = re.sub(dig, '0', s)
    s = s.replace("``", '"')
    s = s.replace("''", '"')
    return s

def construct_token(cols, open_constituents, restrict_features):
    text = cols[3]
    pos = cols[4]
    features = []
    features.append(('word', normalize_text(text)))
    if not restrict_features:
        features.append(('pos', pos))
        features.append(('allCaps', text.isupper()))
        features.append(('upperInitial', text[0].isupper()))

        # character uni-, bi- and trigrams
        for i in range(len(text)):
            features.append(('characters', text[i]))
            if i+1 < len(text):
                features.append(('characters', text[i:i+2]))
            if i+2 < len(text):
                features.append(('characters', text[i:i+3]))

        # constituency features

        constituents_text = cols[5].replace('*', '').replace('-', '')

        while constituents_text.startswith('('):
            constituents_text = constituents_text[1:]
            match = re.match('[A-Z]+', constituents_text)
            constituent = match.group()
            features.append(('constituent start', constituent))
            open_constituents.append(constituent)
            constituents_text = constituents_text[len(constituent):]
        for constituent in open_constituents:
            features.append(('constituent', constituent))
        while constituents_text.startswith(')'):
            constituent = open_constituents.pop()
            features.append(('constituent end', constituent))
            constituents_text = constituents_text[1:]

        assert constituents_text == ''
    
    return Token(text, features)

def load_document(path, span_types=None, restricted_features=False):
    tokens = []
    labels = []
    open_constituents = []
    current_span_type = None
    with open(path) as f:
        for line in f:
            if line.startswith('#'):
                continue
            cols = line.split()
            if len(cols) == 0:
                continue
            tokens.append(construct_token(cols, open_constituents, restricted_features))

            label_col = cols[10]
            if '(' in label_col:
                assert current_span_type is None
                current_span_type = label_col[1:-1]
                assert '(' not in current_span_type
                assert ')' not in current_span_type
                labels.append('B-' + current_span_type)
            else:
                if current_span_type is not None:
                    labels.append('I-' + current_span_type)
                else:
                    labels.append('O')
            if ')' in label_col:
                current_span_type = None
    assert len(open_constituents) == 0
    return Document(tokens, labels)

def load_corpus(path, span_types=None, restricted_features=False):
    documents = []
    for dirpath, dirnames, filenames in os.walk(path):
        for filename in filenames:
            if filename.endswith('.gold_conll'):
                fpath = os.path.join(dirpath, filename)
                documents.append(load_document(fpath, span_types=span_types, restricted_features=restricted_features))
    assert len(documents) > 0
    return Corpus(documents)

if __name__ == '__main__':
    import sys
    corpus_restricted = load_corpus(sys.argv[1], restricted_features=True)
    corpus_full = load_corpus(sys.argv[1])
    breakpoint()
